/*
 *	Sherlock Library -- Linear Representation of Objects
 *
 *	(c) 2005 Martin Mares <mj@ucw.cz>
 *
 *	This software may be freely distributed and used according to the terms
 *	of the GNU Lesser General Public License.
 */

#include "sherlock/sherlock.h"
#include "lib/fastbuf.h"
#include "lib/unaligned.h"
#include "lib/lizard.h"
#include "sherlock/object.h"

#include <stdlib.h>

static uns
obj_linear_size(struct odes *d)
{
  uns sz = 0;
  for (struct oattr *a=d->attrs; a; a=a->next)
    for (struct oattr *b=a; b; b=b->same)
      if (a->attr >= OBJ_ATTR_SON)
	sz += 3 + obj_linear_size(b->son) + 2;
      else
	sz += size_attr(strlen(b->val));
  return sz;
}

static byte *
obj_linear_write(byte *t, struct odes *d)
{
  for (struct oattr *a=d->attrs; a; a=a->next)
    for (struct oattr *b=a; b; b=b->same)
      if (a->attr >= OBJ_ATTR_SON)
	{
	  t = put_attr_push(t, a->attr - OBJ_ATTR_SON);
	  t = obj_linear_write(t, b->son);
	  t = put_attr_pop(t);
	}
      else
	t = put_attr_str(t, a->attr, b->val);
  return t;
}

byte *
obj_linearize(struct odes *d, uns min_compress, uns *plen)
{
  // Create uncompressed linearization
  put_attr_set_type(BUCKET_TYPE_V33);
  uns len = 4 + 1 + obj_linear_size(d);		// starts with 4 bytes of bucket type and a dummy bucket header
  byte *out = xmalloc(len + LIZARD_NEEDS_CHARS);
  PUT_U32(out, BUCKET_TYPE_V33);
  byte *t = put_attr_separator(out+4);
  t = obj_linear_write(t, d);
  ASSERT(t == out+len);

  // Try to compress
  if (min_compress)
    {
      byte *buf_o = out+5;
      uns len_o = len-5;
      uns est_c = 4 + 1 + 8 + len_o * LIZARD_MAX_MULTIPLY + LIZARD_MAX_ADD;
      byte *buf_c = xmalloc(est_c);
      uns len_c = lizard_compress(buf_o, len_o, buf_c+13);
      if (len_c >= len_o * (min_compress / 100.))
	{
	  // Too long
	  xfree(buf_c);
	}
      else
	{
	  PUT_U32(buf_c, BUCKET_TYPE_V33_LIZARD);
	  put_attr_separator(buf_c+4);
	  PUT_U32(buf_c+5, len_o);
	  uns adler = adler32(buf_o, len_o);
	  PUT_U32(buf_c+9, adler);
	  xfree(out);
	  *plen = 13+len_c;
	  return buf_c;
	}
    }

  *plen = len;
  return out;
}

struct odes *
obj_delinearize(struct buck2obj_buf *bbuf, struct mempool *mp, byte *buf, uns len, uns destructive)
{
  struct odes *o = obj_new(mp);
  ASSERT(len >= 4);
  uns buck_type = GET_U32(buf);

  struct fastbuf fb;
  fbbuf_init_read(&fb, buf+4, len-4, destructive);
  if (buck2obj_parse(bbuf, buck_type, len-4, &fb, NULL, NULL, o) < 0)
    return NULL;
  else
    return o;
}
