/*
|| shcp.c - sending large files to multiple destinations
||
|| 2005 (c) Vladimir Jelen <vladimir.jelen@netcentrum.cz>
*/

#include "sherlock/sherlock.h"
#include "lib/conf.h"
#include "lib/lfs.h"
#include "lib/md5.h"
#include "lib/lizard.h"

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/time.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <fcntl.h>
#include <unistd.h>
#include <netinet/in.h>
#include <netdb.h>
#include <string.h>
#include <assert.h>

#define FILE_PERM	0666
#define MAX_ARGS	1024
#define USAGE_HOST	"spatne zavorky"
#define BUFF_SIZE	65536
#define PROTOCOL_V	"SHCP0001"
	     
enum fs_type {
  SHCP_FILE, SHCP_SENDER, SHCP_RECIVER
};

struct arglist {
  int pos;
  char *list[MAX_ARGS];
};

struct host_info {
  int is_local;
  char *host;
  char *file;
  
  struct arglist args;
  
  int buff_pos;
  enum fs_type type;
  
  int dsc;
  int is_w;

  struct hostent *he;
  struct sockaddr_in sa;
};

struct file_info {
  s64 size;
  u32 adler;
  byte key[MD5_HEX_SIZE];
  byte md5[16];
};

static int trace = 1;
static struct file_info fi;

static void shcp_read(struct host_info *hi, void *buff, int len);
static void shcp_write(struct host_info *hi, void *buff, int len);

static void
addarg(struct arglist *a, const char *s)
{
  if(a->pos >= MAX_ARGS)
    die("max args overflow");
  
  a->list[a->pos] = (char *)s;
  a->pos++;
}

static void
open_file(struct host_info *fl, int flags, mode_t mode)
{
  fl->type = SHCP_FILE;
  if((fl->dsc = sh_open(fl->file, flags, mode)) < 0)
    die("Can't open file %s: %m", fl->file); 
}

static void
open_sock(struct host_info *sk, enum fs_type type)
{ 
  socklen_t len;
  sk->type = type;
  sk->sa.sin_family = AF_INET;
  if((sk->dsc = socket(PF_INET, SOCK_STREAM, IPPROTO_TCP)) < 0)
    die("Can't open socket: %m");

  switch(type) {
    case SHCP_RECIVER:
      len = sizeof(sk->sa);
      sk->sa.sin_port = htons(0);
      sk->sa.sin_addr.s_addr = htonl(INADDR_ANY);
      if(bind(sk->dsc, (struct sockaddr *)&(sk->sa), len) < 0)
        die("Bind error: %m");
      if(getsockname(sk->dsc, (struct sockaddr *)&(sk->sa), &len) < 0)
        die("Getsockname error: %m");
      if(listen(sk->dsc, 1) < 0)
        die("Listen error: %m");
      break;

    case SHCP_SENDER:
      sk->he = gethostbyname(sk->host);
      memcpy(&sk->sa.sin_addr, sk->he->h_addr, 4);
      if(connect(sk->dsc, (struct sockaddr *)&(sk->sa), sizeof(sk->sa)) < 0)
        die("Cannot connect to server '%s:%d': %m", sk->host, sk->sa.sin_port);
      break;

    default:
      die("Calling open_sock with type SHCP_FILE");  
  }
}

static void
start_server(struct host_info *hi)
{
  int tmp;
  socklen_t len = sizeof(hi->sa);

  if(hi->type == SHCP_RECIVER) {
    if((tmp = accept(hi->dsc, (struct sockaddr *)&(hi->sa), &len)) < 0)
      die("Accept error: %m");
    close(hi->dsc);
    hi->dsc = tmp;
  }
}

static void
read_headers(struct host_info *hi, struct file_info *fi)
{
  struct stat fs;
  
  ASSERT(hi);
  
  switch(hi->type) {
    case SHCP_FILE:
      if(fstat(hi->dsc, &fs) < 0)
        die("Can't get stats for file %s: %m", hi->file);
      fi->size = fs.st_size;
      break;
      
    case SHCP_RECIVER:
      shcp_read(hi, &fi->size, 8);
      break;

    case SHCP_SENDER:
      die("Can't figure as SHCP_SENDER");  
  }
}

static void
write_headers(struct host_info *hi, struct file_info *fi)
{
  ASSERT(hi);

  switch(hi->type) {
    case SHCP_SENDER:
      shcp_write(hi, &fi->size, 8);
      break;

    case SHCP_FILE:
      break;

    case SHCP_RECIVER:
      die("Can't figure as SHCP_RECIVER");  
  }
}

/*------------------------------------------------------------------------------
|| String functions with limit and \0
*/

static byte *
xstrndup(byte *s, uns n)
{
  uns l = MIN(strlen(s), n);
  return memcpy(xmalloc_zero(l + 1), s, l);
}

/*------------------------------------------------------------------------------
|| Parse host functions
*/

static void
parse_host(struct host_info *hi, char *s) {
  char *pos;
  
  hi->is_local = 0;
  
  pos = strchr(s, ':');
  if(strrchr(s, ':') != pos)
    die("Host '%s' doesn't match pattern [host:]file", s);
  
  if(pos) {
    pos++;
    
    hi->host = xstrndup(s, pos - s - 1);
    hi->file = xstrdup(pos);

    addarg(&(hi->args), "ssh");
    addarg(&(hi->args), hi->host);
    addarg(&(hi->args), "~/run/bin/shcp");
    addarg(&(hi->args), "-l");
    addarg(&(hi->args), hi->file);
  }
  else {
    hi->is_local = 1;
    hi->file = xstrdup(s);
  }
}

static struct host_info **
parse_hosts(int *host_c, int is_reciver, int argc, char **argv) { 
  int args, argl, i, host_i, j;
  struct host_info **hl;
  
  *host_c = host_i = is_reciver;
  argl = args = 0;
  
  for(i = optind; i < argc; i++) {
    if(!strcmp(argv[i], "{")) {
      argl++;
      args = i;
      if(i < (optind + 2 - is_reciver))
        die(USAGE_HOST);
    }
    else if(!strcmp(argv[i], "}")) {
      argl--;
      if(argl < 0 || i == (args + 1))
        die(USAGE_HOST);
    }
    else if(argl == 0)
      (*host_c)++;
  }
  
  if(argl != 0 || *host_c < 2)
    die(USAGE_HOST);
  
  hl = xmalloc(sizeof(struct host_info *) * (*host_c));
  
  for(i = 0; i < *host_c; i++)
    hl[i] = xmalloc(sizeof(struct host_info));
  
  args = 0;
  
  for(i = optind; i < argc; i++) {
    if(!strcmp(argv[i], "{")) {
      if(argl == 0)
        args = i;
      argl++;
    }
    else if(!strcmp(argv[i], "}")) {
      argl--;
      
      if(argl == 0) {
        for(j = args + 1; j < i; j++) {
	  addarg(&(hl[host_i - 1]->args), argv[j]);
        }
      }
    }
    else if(argl == 0){
      parse_host(hl[host_i], argv[i]);
      host_i++;
    }
  }
  
  return hl;
}

static void
open_host(struct host_info *hi, int is_write, int is_reciver)
{
  enum fs_type type = SHCP_SENDER;
  int rw_flags = O_RDONLY | O_DIRECT;

  if((!is_write) && is_reciver)
    type = SHCP_RECIVER;

  if(is_write)
    rw_flags = O_WRONLY | O_CREAT | O_TRUNC | O_DIRECT;
  
  if(hi->is_local)
    open_file(hi, rw_flags, FILE_PERM);
  else
    open_sock(hi, type);
}

static void
fork_hosts(struct host_info **hl, int host_c, int is_reciver)
{
  int i, j;
  int in[2], out[2], res[2], *fdin, *fdout;
  char protocol[8];
  
  for(i = 1; i < host_c; i++) {
    if(!(hl[i]->is_local)) {
      printf("doing cmd: ");
      for(j = 0; j < hl[i]->args.pos; j++)
        printf("%s ", hl[i]->args.list[j]);
      printf("\n");
    
      pipe(res);
      pipe(in);
      pipe(out);

      close(res[0]);
      close(res[1]);
      
      switch(fork()) {
        case -1:
	  die("Fork error: %m");
          break;

        // child
	case 0:
	  dup2(in[0], 0);
	  dup2(out[1], 1);
	  close(in[0]);
	  close(in[1]);
	  close(out[0]);
	  close(out[1]);
	  
	  addarg(&(hl[i]->args), NULL);
	  if(execvp("ssh", hl[i]->args.list) == -1)
	    die("Can't execute ssh: %m");
      }
      
      // parrent
      fdin = &out[0];
      fdout = &in[1];
      close(out[1]);
      close(in[0]);
      read(*fdin, protocol, 8);
      if(memcmp(protocol, PROTOCOL_V, 8))
        die("Different protocol version - server: %s  client: %s" PROTOCOL_V, protocol);
      read(*fdin, &(hl[i]->sa.sin_port), sizeof(hl[i]->sa.sin_port));
      write(*fdout, fi.key, MD5_HEX_SIZE);
      close(out[0]);
      close(in[1]);
    }
  }
  
  open_host(hl[0], 0, is_reciver);
  if(is_reciver) {
    write(1, PROTOCOL_V, 8);
    write(1, &(hl[0]->sa.sin_port), sizeof(hl[0]->sa.sin_port));
    read(0, fi.key, MD5_HEX_SIZE);
    start_server(hl[0]);
  }
  
  for(i = 1; i < host_c; i++)
    open_host(hl[i], i, is_reciver);
}

static void
shcp_read(struct host_info *hi, void *buff, int len) {
  if(hi->type == SHCP_FILE) {
    if(read(hi->dsc, buff, MAX(len, BUFF_SIZE)) < len)
      die("Error reading file %s: %m", hi->file);
  }
  else {
    if(careful_read(hi->dsc, buff, len) < 1)
      die("Error reading socket: %m");
  }
}

static void
shcp_write(struct host_info *hi, void *buff, int len) {
  if(hi->type == SHCP_FILE) {
    if(write(hi->dsc, buff, MAX(len, BUFF_SIZE)) < MAX(len, BUFF_SIZE))
      die("Error writing file %s: %m", hi->file);
  }
  else {
    if(careful_write(hi->dsc, buff, len) < 1)
      die("Error writing socket: %m/");
  }
}

int
main(int argc, char **argv)
{
  char *buff;
  int i, opt, is_reciver, host_c, status, retval;
  s64 size;
  struct host_info **host_list;
  struct MD5Context md5c;

  is_reciver = size = 0;
  
  /*----------------------------------------------------------------------------
  || Parsing & seting options
  */
  
  while((opt = getopt(argc, argv, "ls")) != -1) {
    switch(opt) {
      case 'l':
        is_reciver = 1;
	break;
      
      case 's':
        trace = 0;
	break;
	
      default:
        die("shcp [-s] [-l] [input_file] [host:]file ... [{ [host:]file }]");
    }
  }

  if(!is_reciver) {
    byte keybin[MD5_SIZE];
    randomkey(keybin, MD5_SIZE);
    md5_to_hex(keybin, fi.key);
  }

  host_list = parse_hosts(&host_c, is_reciver, argc, argv);
  fork_hosts(host_list, host_c, is_reciver);

  if(!(buff = valloc(BUFF_SIZE)))
    die("Can't allocate %d bytes of memory: %m");
  
  read_headers(host_list[0], &fi);
  for(i = 1; i < host_c; i++)
    write_headers(host_list[i], &fi);

  
  fi.adler = update_adler32(0, NULL, 0);
  size = fi.size;
  while(size > 0) {
    int sz = BUFF_SIZE;
    if(size < BUFF_SIZE)
      sz = size;
   
    shcp_read(host_list[0], buff, sz);	    
    for(i = 1; i < host_c; i++)
      shcp_write(host_list[i], buff, sz);

    fi.adler = update_adler32(fi.adler, buff, sz);
    size -= sz;
  }
  
  MD5Init(&md5c);
  MD5Update(&md5c, fi.key, strlen(fi.key));
  MD5Update(&md5c, (byte *) &fi.adler, sizeof(fi.adler));
  MD5Final(fi.md5, &md5c);

  if(is_reciver) {
    byte sig[16];
    shcp_read(host_list[0], sig, sizeof(sig));
    if(memcmp(fi.md5, sig, sizeof(sig)) != 0)
      die("File checksum mismatch");
  }
  
  for(i = 1; i < host_c; i++) {
    if(host_list[i]->type == SHCP_FILE)
      sh_ftruncate(host_list[i]->dsc, fi.size);
    else
      shcp_write(host_list[i], fi.md5, sizeof(fi.md5));
  }
  
  
  // code something better - which child exited with error etc.
  retval = 0;
  while((i = wait(&status)) > 0) {
    if(status) {
      retval = 1;
      log(L_ERROR, "Child process error");
    }
  }
  
  return retval;
}
